import cbor2
from pydantic import BaseModel

from local_migrator import cbor_decoder, cbor_encoder


class SampleModel(BaseModel):
    field1: int
    field2: str


data = SampleModel(field1=4, field2="abc")

with open("sample.cbor", "wb") as f_p:
    cbor2.dump(data, f_p, default=cbor_encoder)

with open("sample.cbor", "rb") as f_p2:
    data2 = cbor2.load(f_p2, object_hook=cbor_decoder)

assert data == data2



"""
This module contains utility for registration migration information for class.
"""

import importlib
import inspect
import warnings
from dataclasses import dataclass
from functools import wraps
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, TypeVar, Union

from packaging.version import Version
from packaging.version import parse as parse_version

MRO_NO_SUPERCLASS = 2


def class_to_str(cls) -> str:
    """Get full qualified name for a given class."""
    if cls.__module__.startswith("pydantic.dataclass"):
        cls = cls.__mro__[1]
        return class_to_str(cls)
    return f"{cls.__module__}.{cls.__qualname__}"


def get_super_class(cls: Type) -> Optional[Type]:
    """Get parent class for a given class"""
    if len(cls.__mro__) < MRO_NO_SUPERCLASS:
        return None
    if cls.__module__.startswith("pydantic.dataclass"):
        # Not covered by tests, because I cannot reproduce problem that causes this.
        # It may be connected with pydantic version
        return get_super_class(cls.__mro__[1])
    return cls.__mro__[1]


def str_to_version(version: Union[str, Version]) -> Version:
    """If version passed as sting then convert it to Version object, otherwise return untouched."""
    return parse_version(version) if isinstance(version, str) else version


def _class_str_replace(func):
    @wraps(func)
    def _wrap(*args, **kwargs):
        if "class_str" in kwargs:
            warnings.warn("class_str argument is deprecated, use cls instead", FutureWarning, stacklevel=2)
            kwargs["cls"] = kwargs.pop("class_str")
        return func(*args, **kwargs)

    return _wrap


MigrationCallable = Callable[[Dict[str, Any]], Dict[str, Any]]
MigrationInfo = Tuple[Version, MigrationCallable]
"""Type describing single migration entry. For given class Version number should be unique."""
MigrationStartInfo = Tuple[Union[str, Version], MigrationCallable]

T = TypeVar("T")

RegisterReturnType = Union[Type[T], Callable[[Type[T]], Type[T]]]


@dataclass(frozen=True)
class TypeInfo:
    """
    Class for storing information in :py:class:`~.MigrationRegistration`

    :ivar str base_path: full qualified path to current class module and name
    :ivar typing.Type type_: class itself
    :ivar packaging.version.Version version: current clas version
    :ivar typing.List[~.MigrationInfo] migrations: list of migrations for deserialize old version
    :ivar bool use_parent_migrations: if migrations from parent class should be applied when deserialized object.
    """

    base_path: str
    type_: Type
    version: Version
    migrations: List[MigrationInfo]
    use_parent_migrations: bool
    allow_errors_in_values: bool


class MigrationRegistration:
    """
    Implementation of class register to storage information needed for migration from previous version.
    """

    def __init__(self):
        self._data_dkt: Dict[str, TypeInfo] = {}

    def register(  # noqa: PLR0913
        self,
        cls: Optional[Type] = None,
        version: Union[str, Version] = "0.0.0",
        migrations: Optional[List[MigrationStartInfo]] = None,
        old_paths: Optional[List[str]] = None,
        use_parent_migrations: bool = True,
        allow_errors_in_values: bool = False,
    ) -> RegisterReturnType:
        """
        Register class instance for storage information needed for deserialization of object from older version.


        :param cls: class to be registered
        :param version: current version of class
        :param  typing.List[MigrationInfo] migrations: list of migrations for deserialize old version
        :param old_paths: old name of class with modules
        :param use_parent_migrations: if migrations from parent class should be applied when deserialized object
        :param allow_errors_in_values: if errors in constructor kwargs should be ignored. Added to not block creating
            of custom Mapping class that could contain broken items.
        :return: class itself if cls parameter is provided. Otherwise,
            one argument function which will consume Type to be registered.
        """
        migrations = [] if migrations is None else sorted((str_to_version(x), y) for x, y in migrations)
        if old_paths is None:
            old_paths = []
        version = str_to_version(version)

        if migrations and max(x for x, _ in migrations) > version:
            raise ValueError("class version lower than in migrations")

        def _register(cls_):
            base_path = class_to_str(cls_)
            type_info = TypeInfo(
                base_path=base_path,
                type_=cls_,
                version=version,
                migrations=migrations,
                use_parent_migrations=use_parent_migrations,
                allow_errors_in_values=allow_errors_in_values,
            )
            if base_path in self._data_dkt:
                raise RuntimeError(f"Class name {base_path} already taken by {self._data_dkt[base_path].base_path}")
            self._data_dkt[base_path] = type_info
            for name in old_paths:
                if name in self._data_dkt and self._data_dkt[name].base_path != base_path:
                    raise RuntimeError(f"Class name {name} already taken by {self._data_dkt[name].base_path}")
                self._data_dkt[name] = type_info
            return cls_

        return _register if cls is None else _register(cls)

    def use_parent_migrations(self, name: str) -> bool:
        """
        Check if parent migrations should be used.

        :param name: full qualified path to class
        :return: information if parent class migrations should be applied.
        """
        self._register_missed(class_str=name)
        return self._data_dkt[name].use_parent_migrations

    def get_version(self, cls: Type) -> Version:
        """For a given class return version with which given class was registered using :py:meth:`register`"""
        class_str = class_to_str(cls)
        self._register_missed(class_str=class_str)
        return self._data_dkt[class_str].version

    def get_class(self, class_str: str) -> Type:
        """
        Get class base of qualified name. Could be done using current or old path.

        Qualified name is determined using :py:func:`class_to_str` and old path comes from the ``old_paths`` argument
        of :py:meth:`register` method.
        """
        self._register_missed(class_str=class_str)
        return self._data_dkt[class_str].type_

    def allow_errors_in_values(self, cls: Union[str, Type]) -> bool:
        """
        Check if class should allow errors in values.

        :param cls: class or full qualified path to class
        """
        if not isinstance(cls, str):
            cls = class_to_str(cls)
        self._register_missed(class_str=cls)
        return self._data_dkt[cls].allow_errors_in_values

    @_class_str_replace
    def migrate_data(
        self, cls: Union[str, Type], class_str_to_version_dkt: Dict[str, Union[str, Version]], data: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Apply migrations base on register state. Current implementation does not support multiple inheritance.

        :param cls: fully qualified class path
        :param class_str_to_version_dkt: for each parent class information about version during serialization.
            If class is absent from this dict then assumed version is "0.0.0"
        :param data: dict of kwargs to constructor of class
        """
        if not isinstance(cls, str):
            cls = class_to_str(cls)

        if self.use_parent_migrations(cls):
            super_klass = get_super_class(self.get_class(cls))
            if super_klass is not None:
                data = self.migrate_data(class_to_str(super_klass), class_str_to_version_dkt, data)
        version = str_to_version(class_str_to_version_dkt.get(cls, "0.0.0"))
        for version_, migration in self._data_dkt[cls].migrations:
            if version < version_:
                data = migration(data)
        return data

    def _register_missed(self, class_str):
        """Register class if missed from register"""
        if class_str in self._data_dkt:
            return
        module_name, class_name = class_str.rsplit(".", maxsplit=1)
        class_path = [class_name]
        while True:
            try:
                module = importlib.import_module(module_name)
                break
            except ModuleNotFoundError as e:
                module_name_split = module_name.rsplit(".", maxsplit=1)
                if len(module_name_split) == 1:
                    raise ValueError(f"Class {class_str} not found") from e

                module_name, class_name_ = module_name_split
                class_path.append(class_name_)
        if class_str in self._data_dkt:
            return
        class_ = module
        try:
            for name in class_path[::-1]:
                class_ = getattr(class_, name)
        except AttributeError as e:
            raise ValueError(f"Class {class_str} not found") from e
        self.register(class_)


# The global instance of register is use because registration is performed on import time.
# There should no information storage for objects.
REGISTER = MigrationRegistration()
"""Default register to storage class information. Instance of :py:class:`MigrationRegistration`"""


def rename_key(from_key: str, to_key: str, optional=False) -> MigrationCallable:
    """
    simple migration function for rename fields

    :param from_key: original name
    :param to_key: destination name
    :param optional: if migration is required (for backward compatibility)
    :return: migration function
    """

    def _migrate(dkt: Dict[str, Any]) -> Dict[str, Any]:
        if optional and from_key not in dkt:
            return dkt
        res_dkt = dkt.copy()
        res_dkt[to_key] = res_dkt.pop(from_key)
        return res_dkt

    return _migrate


def update_argument(argument_name):
    """
    This is decorator for move conversion of dict to class outside function code.
    It first inspects function signature to determine type th which argument should be converted.
    Then, if argument is passed as dict then all migrations from :py:attr:`REGISTER` all applied,
    then object is constructed and replace base one.

    :param argument_name: name of argument which should be converted

    Example::

        @register_class(version="0.0.1", migrations=[("0.0.1", rename_key("value", "value1"))])
        class DataClass:
            def __init__(value1, value2)
                self.value1 = value1
                self.value2 = value2

        @update_argument("arg")
        def some_func(arg: DataClass):
            print(arg.value1)

        some_func({"value": 1, "value2": 5})

    """

    def _wrapper(func):
        signature = inspect.signature(func)
        if argument_name not in signature.parameters:  # pragma: no cover
            raise RuntimeError("Argument should be accessible using inspect module.")
        arg_index = list(signature.parameters).index(argument_name)
        klass = signature.parameters[argument_name].annotation
        if not inspect.isclass(klass):  # pragma: no cover
            raise ValueError(f"Annotation {klass} of {argument_name} parameter is not a class")

        @wraps(func)
        def _update_from_dict(*args, **kwargs):
            if argument_name in kwargs and isinstance(kwargs[argument_name], dict):
                kwargs = kwargs.copy()
                kw = REGISTER.migrate_data(class_to_str(klass), {}, kwargs[argument_name])
                kwargs[argument_name] = klass(**kw)
            elif len(args) > arg_index and isinstance(args[arg_index], dict):
                args = list(args)
                kw = REGISTER.migrate_data(class_to_str(klass), {}, args[arg_index])
                args[arg_index] = klass(**kw)
            return func(*args, **kwargs)

        return _update_from_dict

    return _wrapper


def register_class(  # noqa: PLR0913
    cls: Optional[Type[T]] = None,
    version: Union[str, Version] = "0.0.0",
    migrations: Optional[List[MigrationStartInfo]] = None,
    old_paths: Optional[List[str]] = None,
    use_parent_migrations: bool = True,
    allow_errors_in_values: bool = False,
) -> RegisterReturnType:
    """
    This is wrapper for call :py:meth:`MigrationRegistration.register` of default register instance.
    Please see its documentation for details.

    :param cls: class to be registered
    :param version: current version of class
    :param  typing.List[MigrationInfo] migrations: list of migrations for deserialize old version
    :param old_paths: old name of class with modules
    :param use_parent_migrations: if migrations from parent class should be applied when deserialized object
    :param allow_errors_in_values: if errors in constructor kwargs should be ignored. Added to not block creating
        of custom Mapping class that could contain broken items.
    :return: class itself if cls parameter is provided. Otherwise,
        one argument function which will consume Type to be registered.

    Examples::

        @register_class(version="0.0.1", migrations=[("0.0.1", rename_key("value", "value1"))])
        class DataClass:
            def __init__(value1, value2)
                self.value1 = value1
                self.value2 = value2

    or::


        class DataClass2:
            def __init__(value1, value2)
                self.value1 = value1
                self.value2 = value2

        register_class(DataClass2, version="0.0.1", migrations=[("0.0.1", rename_key("value", "value1"))])

    """
    return REGISTER.register(cls, version, migrations, old_paths, use_parent_migrations, allow_errors_in_values)



import dataclasses
import enum
import json
import typing
from pathlib import Path

from ._class_register import REGISTER, class_to_str

try:
    from pydantic import BaseModel
except ImportError:  # pragma: no cover
    # allow to use in environment without pydantic.
    class BaseModel:  # type: ignore  [no-redef]
        pass


try:
    from pydantic.v1 import BaseModel as BaseModelV1
except ImportError:

    class BaseModelV1:  # type: ignore  [no-redef]
        pass


try:
    from numpy import floating, integer, ndarray
except ImportError:  # pragma: no cover
    # allow to use in environment without numpy.
    class ndarray:  # type: ignore  [no-redef]
        pass

    class integer:  # type: ignore  [no-redef]
        pass

    class floating:  # type: ignore  [no-redef]
        pass


def add_class_info(obj: typing.Any, dkt: dict) -> dict:
    return {
        "__class__": class_to_str(obj.__class__),
        "__class_version_dkt__": {
            class_to_str(sup_obj): str(REGISTER.get_version(sup_obj))
            for sup_obj in obj.__class__.__mro__
            if class_to_str(sup_obj)
            not in {
                "object",
                "pydantic.main.BaseModel",
                "pydantic.utils.Representation",
                "enum.Enum",
                "builtins.object",
                "typing.Generic",
            }
            and not class_to_str(sup_obj).startswith("collections.abc")
        },
        "__values__": dkt,
    }


def object_encoder(obj: typing.Any):  # noqa: PLR0911
    """
    Function changing supported types to basic python types supported by most
    serializers and which could be restored by :py:func:`nme_object_hook` function.

    Supported types are:

    * :py:class:`enum.Enum`
    * :py:func:`dataclasses.dataclass`
    * :py:class:`numpy.ndarray`
    * :py:class:`pydantic.BaseModel`
    * :py:class:`numpy.integer` (change to pure int)
    * :py:class:`numpy.floating` (change to pure float)
    * :py:class:`pathlib.Path` (Serialized to string)
    * Any class with an ``as_dict`` method. This method should return a dictionary of valid constructor arguments.

    :param obj: object to be encoded.
    :return: encoded object for supported types. Otherwise ``None``.

    """
    if isinstance(obj, enum.Enum):
        dkt = {"value": obj.value}
        return add_class_info(obj, dkt)
    if dataclasses.is_dataclass(obj):
        fields = dataclasses.fields(obj)
        dkt = {x.name: getattr(obj, x.name) for x in fields}
        return add_class_info(obj, dkt)

    if isinstance(obj, ndarray):
        return obj.tolist()

    if isinstance(obj, (BaseModel, BaseModelV1)):
        try:
            dkt = dict(obj)
        except (ValueError, TypeError):
            dkt = obj.dict()  # workaround for napari Colormap class
        return add_class_info(obj, dkt)

    if hasattr(obj, "as_dict"):
        dkt = obj.as_dict()
        return add_class_info(obj, dkt)

    if isinstance(obj, integer):
        return int(obj)
    if isinstance(obj, floating):
        return float(obj)
    if isinstance(obj, Path):
        return str(obj)
    return None


class Encoder(json.JSONEncoder):
    """
    JSONEncoder subclass for serializing Python objects into JSON.
    For list of supported types check :py:func:`nme_object_encoder` function.
    """

    def default(self, o):
        """
        Implementation that calls :py:func:`nme_object_encoder` function.
        """
        val = object_encoder(o)
        if val is None:  # pragma: no cover
            return super().default(o)
        return val


def check_for_errors_in_dkt_values(dkt: dict) -> typing.List[str]:
    """
    Function checking if any of values in dict contains ``"__error__"`` key.

    :param dkt: dictionary to check.
    :return: list of keys that value is dict containing ``"__error__"`` key.
    """
    return [key for key, value in dkt.items() if isinstance(value, dict) and "__error__" in value]


def object_hook(dkt: dict) -> typing.Any:
    """
    Function restoring supported types from :py:func:`nme_object_encoder` function output.

    If ``dkt`` does not contain ``__class__`` key, it is returned as is.

    If the restoring object fails then function return dict with ``"__error__"`` key.

    :param dkt: dictionary with data to restore.
    """
    if "__error__" in dkt:
        dkt.pop("__error__")  # different environments without same plugins installed
    if "__class__" in dkt:
        if "__values__" not in dkt:
            cls_str = dkt.pop("__class__")
            version_dkt = dkt.pop("__class_version_dkt__") if "__class_version_dkt__" in dkt else {cls_str: "0.0.0"}
            dkt = {"__values__": dkt, "__class__": cls_str, "__class_version_dkt__": version_dkt}
        try:
            cls = REGISTER.get_class(dkt["__class__"])
        except (KeyError, ValueError):
            dkt["__error__"] = f"Class {dkt['__class__']} not found in register."
            return dkt
        problematic_fields = check_for_errors_in_dkt_values(dkt["__values__"])
        if problematic_fields and not REGISTER.allow_errors_in_values(cls):
            dkt["__error__"] = f"Error in fields: {', '.join(problematic_fields)}"
            return dkt
        try:
            dkt_migrated = REGISTER.migrate_data(dkt["__class__"], dkt["__class_version_dkt__"], dkt["__values__"])
            cls = REGISTER.get_class(dkt["__class__"])
            return cls(**dkt_migrated)
        except Exception as e:  # pylint: disable=W0703
            dkt["__error__"] = str(e)

    return dkt


nme_object_hook = object_hook
nme_object_encoder = object_encoder
NMEEncoder = Encoder



# pylint: disable=R0201

import json
from enum import Enum
from pathlib import Path

import numpy as np
import pytest
from pydantic import BaseModel, Extra, dataclasses

from local_migrator import Encoder, class_to_str, object_hook, register_class, rename_key
from local_migrator._serialize_hooks import add_class_info

try:
    from napari.utils import Colormap
    from napari.utils.notifications import NotificationSeverity
except ImportError:
    Colormap = None
    NotificationSeverity = None


@dataclasses.dataclass
class SampleDataclass:
    filed1: int
    field2: str


class SamplePydantic(BaseModel):
    sample_int: int
    sample_str: str
    sample_dataclass: SampleDataclass


class SampleAsDict:
    def __init__(self, value1, value2):
        self.value1 = value1
        self.value2 = value2

    def as_dict(self):
        return {"value1": self.value1, "value2": self.value2}


class RadiusType(Enum):
    NO = 0
    R2D = 1
    R3D = 2


@pytest.mark.skipif(Colormap is None, reason="napari not installed")
def test_colormap_dump(tmp_path):
    cmap_list = [Colormap([(0, 0, 0), (1, 1, 1)]), Colormap([(0, 0, 0), (1, 1, 1)], controls=[0, 1])]
    with open(tmp_path / "test.json", "w") as f_p:
        json.dump(cmap_list, f_p, cls=Encoder)

    with open(tmp_path / "test.json") as f_p:
        cmap_list2 = json.load(f_p, object_hook=object_hook)

    assert np.array_equal(cmap_list[0].colors, cmap_list2[0].colors)
    assert np.array_equal(cmap_list[0].controls, cmap_list2[0].controls)
    assert np.array_equal(cmap_list[1].colors, cmap_list2[1].colors)
    assert np.array_equal(cmap_list[1].controls, cmap_list2[1].controls)

    cmap_list = [
        Colormap([(0, 0, 0), (1, 1, 1)]),
        Colormap([(0, 0, 0), (0, 0, 0), (1, 1, 1), (1, 1, 1)], controls=[0, 0.1, 0.8, 1]),
    ]
    with open(tmp_path / "test2.json", "w") as f_p:
        json.dump(cmap_list, f_p, cls=Encoder)

    with open(tmp_path / "test2.json") as f_p:
        cmap_list2 = json.load(f_p, object_hook=object_hook)

    assert np.array_equal(cmap_list[0].colors, cmap_list2[0].colors)
    assert np.array_equal(cmap_list[0].controls, cmap_list2[0].controls)
    assert np.array_equal(cmap_list[1].colors, cmap_list2[1].colors)
    assert np.array_equal(cmap_list[1].controls, cmap_list2[1].controls)
    assert cmap_list2[1].controls[0] == 0
    assert cmap_list2[1].controls[-1] == 1
    assert np.array_equal(cmap_list[1].colors[0], cmap_list2[1].colors[0])
    assert np.array_equal(cmap_list[1].colors[-1], cmap_list2[1].colors[-1])


@pytest.mark.parametrize("dtype", [np.uint8, np.uint16, np.uint32, np.float32, np.float64])
def test_dump_numpy_types(dtype):
    data = {"a": dtype(2)}
    text = json.dumps(data, cls=Encoder)
    loaded = json.loads(text)
    assert loaded["a"] == 2


class TestNMEEncoder:
    @pytest.mark.skipif(NotificationSeverity is None, reason="napari not installed")
    def test_enum_serialize(self, tmp_path):
        data = {"value1": RadiusType.R2D, "value2": RadiusType.NO, "value3": NotificationSeverity.ERROR}
        with (tmp_path / "test.json").open("w") as f_p:
            json.dump(data, f_p, cls=Encoder)
        with (tmp_path / "test.json").open("r") as f_p:
            data2 = json.load(f_p, object_hook=object_hook)
        assert data2["value1"] == RadiusType.R2D
        assert data2["value2"] == RadiusType.NO
        assert data2["value3"] == NotificationSeverity.ERROR

    def test_dataclass_serialze(self, tmp_path):
        data = {"value": SampleDataclass(1, "text")}
        with (tmp_path / "test.json").open("w") as f_p:
            json.dump(data, f_p, cls=Encoder)
        with (tmp_path / "test.json").open("r") as f_p:
            data2 = json.load(f_p, object_hook=object_hook)

        assert isinstance(data2["value"], SampleDataclass)
        assert data2["value"] == SampleDataclass(1, "text")

    @pytest.mark.skipif(Colormap is None, reason="napari not installed")
    def test_pydantic_serialize(self, tmp_path):
        data = {
            "color1": Colormap(colors=[[0, 0, 0], [0, 0, 0]], controls=[0, 1]),
            "other": SamplePydantic(sample_int=1, sample_str="text", sample_dataclass=SampleDataclass(1, "text")),
        }
        with (tmp_path / "test.json").open("w") as f_p:
            json.dump(data, f_p, cls=Encoder)
        with (tmp_path / "test.json").open("r") as f_p:
            data2 = json.load(f_p, object_hook=object_hook)
        assert data2["color1"] == Colormap(colors=[[0, 0, 0], [0, 0, 0]], controls=[0, 1])
        assert isinstance(data2["other"], SamplePydantic)
        assert isinstance(data2["other"].sample_dataclass, SampleDataclass)

    def test_numpy_serialize(self, tmp_path):
        data = {"arr": np.arange(10), "f": np.float32(0.1), "i": np.int16(1000)}
        with (tmp_path / "test.json").open("w") as f_p:
            json.dump(data, f_p, cls=Encoder)
        with (tmp_path / "test.json").open("r") as f_p:
            data2 = json.load(f_p, object_hook=object_hook)
        assert data2["arr"] == list(range(10))
        assert np.isclose(data["f"], 0.1)
        assert data2["i"] == 1000

    def test_class_with_as_dict(self, tmp_path):
        data = {"d": SampleAsDict(1, 10)}
        with (tmp_path / "test.json").open("w") as f_p:
            json.dump(data, f_p, cls=Encoder)
        with (tmp_path / "test.json").open("r") as f_p:
            data2 = json.load(f_p, object_hook=object_hook)
        assert isinstance(data2["d"], SampleAsDict)
        assert data2["d"].value1 == data["d"].value1
        assert data2["d"].value2 == data["d"].value2

    def test_sub_class_serialization(self, tmp_path):
        ob = DummyClassForTest.DummySubClassForTest(1, 2)
        with (tmp_path / "test.json").open("w") as f_p:
            json.dump(ob, f_p, cls=Encoder)
        with (tmp_path / "test.json").open("r") as f_p:
            ob2 = json.load(f_p, object_hook=object_hook)
        assert ob2.data1 == 1
        assert ob2.data2 == 2


def test_add_class_info_pydantic(clean_register):
    @register_class
    class SampleClass(BaseModel):
        field: int = 1

    dkt = {}
    dkt = add_class_info(SampleClass(), dkt)
    assert "__class__" in dkt
    assert class_to_str(SampleClass) == dkt["__class__"]
    assert len(dkt["__class_version_dkt__"]) == 1
    assert dkt["__class_version_dkt__"][class_to_str(SampleClass)] == "0.0.0"


def test_add_class_info_enum(clean_register):
    @register_class
    class SampleEnum(Enum):
        field = 1

    dkt = {}
    dkt = add_class_info(SampleEnum.field, dkt)
    assert "__class__" in dkt
    assert class_to_str(SampleEnum) == dkt["__class__"]
    assert len(dkt["__class_version_dkt__"]) == 1
    assert dkt["__class_version_dkt__"][class_to_str(SampleEnum)] == "0.0.0"


def test_path_serialization(tmp_path):
    with (tmp_path / "test.json").open("w") as f_p:
        json.dump(Path(), f_p, cls=Encoder)
    with (tmp_path / "test.json").open("r") as f_p:
        data = json.load(f_p, object_hook=object_hook)
    assert str(Path()) == data


def test_error_deserialization(clean_register, tmp_path):
    class SampleEnum(Enum):
        field = 1

    with (tmp_path / "test.json").open("w") as f_p:
        f_p.write(
            '{"value": 1, "__class__": "test_json_hooks.test_error_deserialization.<locals>.SampleEnum",'
            '"__class_version_dkt__": {"test_json_hooks.test_error_deserialization.<locals>.SampleEnum": "0.0.0"}}'
        )
    with (tmp_path / "test.json").open("r") as f_p:
        data2 = json.load(f_p, object_hook=object_hook)
    assert "__error__" in data2

    register_class(SampleEnum)

    with (tmp_path / "test2.json").open("w") as f_p:
        json.dump(data2, f_p, cls=Encoder)

    with (tmp_path / "test2.json").open("r") as f_p:
        data3 = json.load(f_p, object_hook=object_hook)

    assert data3 == SampleEnum.field


class TestNMEObjectHook:
    def test_no_inheritance_read(self, clean_register, tmp_path):
        @register_class(version="0.0.1", migrations=[("0.0.1", rename_key("field", "field1"))])
        class BaseClass(BaseModel):
            field1: int = 1

        @register_class(
            version="0.0.1", migrations=[("0.0.1", rename_key("field", "field1"))], use_parent_migrations=False
        )
        class MainClass(BaseClass):
            field2: int = 5

        data_str = """
        {"field": 1, "field2": 5,
         "__class__":
         "test_json_hooks.TestNMEObjectHook.test_no_inheritance_read.<locals>.MainClass",
         "__class_version_dkt__": {
         "test_json_hooks.TestNMEObjectHook.test_no_inheritance_read.<locals>.MainClass": "0.0.0",
         "test_json_hooks.TestNMEObjectHook.test_no_inheritance_read.<locals>.BaseClass": "0.0.0"
         }}
         """

        ob = json.loads(data_str, object_hook=object_hook)
        assert isinstance(ob, MainClass)

    def test_error_in_object_restore(self, clean_register):
        @register_class
        class SubClass(BaseModel, extra=Extra.forbid):
            field: int = 1

        @register_class
        class MainClass(BaseModel):
            field: int = 1
            sub11: SubClass = SubClass()

        data_str = """
        {"__class__": "test_json_hooks.TestNMEObjectHook.test_error_in_object_restore.<locals>.MainClass",
        "__class_version_dkt__":
        {"test_json_hooks.TestNMEObjectHook.test_error_in_object_restore.<locals>.MainClass": "0.0.0"},
        "__values__": {"field": 1,
        "sub11": {"__class__": "test_json_hooks.TestNMEObjectHook.test_error_in_object_restore.<locals>.SubClass",
        "__class_version_dkt__":
        {"test_json_hooks.TestNMEObjectHook.test_error_in_object_restore.<locals>.SubClass": "0.0.0"},
        "__values__": {"field": 1, "eee": 1}}}}
        """

        ob = json.loads(data_str, object_hook=object_hook)
        assert isinstance(ob, dict)
        assert "__error__" in ob
        assert ob["__error__"] == "Error in fields: sub11"


class DummyClassForTest:
    class DummySubClassForTest:
        def __init__(self, data1, data2):
            self.data1, self.data2 = data1, data2

        def as_dict(self):
            return {"data1": self.data1, "data2": self.data2}



from dataclasses import dataclass
from enum import Enum

import cbor2
import pytest
from pydantic import BaseModel

from local_migrator import cbor_decoder, cbor_encoder, class_to_str, register_class, rename_key


class RadiusType(Enum):
    NO = 0
    R2D = 1
    R3D = 2


@dataclass
class SampleDataclass:
    filed1: int
    field2: str


class SamplePydantic(BaseModel):
    sample_int: int
    sample_str: str
    sample_dataclass: SampleDataclass


class SampleAsDict:
    def __init__(self, value1, value2):
        self.value1 = value1
        self.value2 = value2

    def as_dict(self):
        return {"value1": self.value1, "value2": self.value2}

    def __eq__(self, other):
        return self.value1 == other.value1 and self.value2 == other.value2

    def __repr__(self) -> str:
        return f"SampleAsDict(value1={self.value1}, value2={self.value2})"


def test_simple(tmp_path):
    data = {"aa": 1, "bb": 2}
    with open(tmp_path / "test.cbor", "wb") as f_p:
        cbor2.dump(data, f_p, default=cbor_encoder)

    with open(tmp_path / "test.cbor", "rb") as f_p:
        data2 = cbor2.load(f_p, object_hook=cbor_decoder)
    assert data2 == data


def test_hook_failure(tmp_path):
    class DummyClass:
        def __init__(self):
            pass

    with pytest.raises(TypeError), open(tmp_path / "test.cbor", "wb") as f_p:
        cbor2.dump(DummyClass(), f_p, default=cbor_encoder)


def test_serialize_enum(tmp_path, clean_register):
    with open(tmp_path / "test.cbor", "wb") as f_p:
        cbor2.dump(RadiusType.NO, f_p, default=cbor_encoder)

    with open(tmp_path / "test.cbor", "rb") as f_p:
        data = cbor2.load(f_p, object_hook=cbor_decoder)
    assert data == RadiusType.NO


def test_dataclass_serialize(tmp_path, clean_register):
    data = SampleDataclass(1, "cc")
    with open(tmp_path / "test.cbor", "wb") as f_p:
        cbor2.dump(data, f_p, default=cbor_encoder)

    with open(tmp_path / "test.cbor", "rb") as f_p:
        data2 = cbor2.load(f_p, object_hook=cbor_decoder)
    assert data2 == data


def test_pydantic_serialize(tmp_path, clean_register):
    data = SamplePydantic(sample_int=2, sample_str="ee", sample_dataclass=SampleDataclass(1, "cc"))
    with open(tmp_path / "test.cbor", "wb") as f_p:
        cbor2.dump(data, f_p, default=cbor_encoder)

    with open(tmp_path / "test.cbor", "rb") as f_p:
        data2 = cbor2.load(f_p, object_hook=cbor_decoder)
    assert data2 == data


def test_as_dict_serialize(tmp_path, clean_register):
    data = SampleAsDict(value1=1, value2=[1, 2, 3])
    with open(tmp_path / "test.cbor", "wb") as f_p:
        cbor2.dump(data, f_p, default=cbor_encoder)

    with open(tmp_path / "test.cbor", "rb") as f_p:
        data2 = cbor2.load(f_p, object_hook=cbor_decoder)
    assert data2 == data


def test_migration_base(tmp_path, clean_register):
    @register_class
    class SampleClass(BaseModel):
        field: int = 1

    with open(tmp_path / "test.cbor", "wb") as f_p:
        cbor2.dump(SampleClass(), f_p, default=cbor_encoder)

    clean_register()

    def provide_default(dkt):
        dkt["field2"] = 7
        return dkt

    @register_class(
        version="0.0.2",
        old_paths=[class_to_str(SampleClass)],
        migrations=[("0.0.1", rename_key("field", "field1")), ("0.0.2", provide_default)],
    )
    class SampleClass2(BaseModel):
        field1: int = 1
        field2: int = 4

    with open(tmp_path / "test.cbor", "rb") as f_p:
        data = cbor2.load(f_p, object_hook=cbor_decoder)
    assert SampleClass2(field1=1, field2=7) == data



from importlib import metadata

from ._class_register import (
    REGISTER,
    MigrationInfo,
    MigrationRegistration,
    class_to_str,
    register_class,
    rename_key,
    update_argument,
)
from ._serialize_hooks import Encoder, check_for_errors_in_dkt_values, object_encoder, object_hook
from .version import version as __version__

try:
    nme_version = metadata.version("nme")
except metadata.PackageNotFoundError:
    pass
else:  # pragma: no cover
    from packaging.version import parse

    if parse(nme_version) <= parse("0.1.6"):
        raise ImportError("local_migrator is incompatible with nme<=0.1.6. You need to upgrade or uninstall nme.")
    del parse
    del nme_version

del metadata


def cbor_encoder(encoder, value):
    """
    Cbor encoder hook. Use :py:func:`nme_object_encoder` to encode objects.

    :param encoder: cbor2.Encoder
    :param value: object to be encoded

    Examples::

        with open(path_to_file, "wb") as f_p:
            cbor2.dump(data, f_p, default=nme_cbor_encoder)
    """
    res = object_encoder(value)
    if res is None:
        raise TypeError(f"Cannot encode {value} of class {type(value)}")
    return encoder.encode(res)


def cbor_decoder(decoder, value):  # noqa: ARG001
    """
    Cbor decoder hook. Use :py:func:`nme_object_hook` to decode objects.

    :param decoder: cbor2.Decoder
    :param value: object to be decoded

    Examples::

        with open(path_to_file, "rb") as f_p:
            data = cbor2.load(f_p, object_hook=nme_cbor_decoder)

    """
    return object_hook(value)


nme_object_hook = object_hook
NMEEncoder = Encoder
nme_cbor_encoder = cbor_encoder
nme_cbor_decoder = cbor_decoder


__all__ = (
    "class_to_str",
    "check_for_errors_in_dkt_values",
    "register_class",
    "object_hook",
    "nme_object_hook",
    "rename_key",
    "MigrationInfo",
    "MigrationRegistration",
    "Encoder",
    "NMEEncoder",
    "REGISTER",
    "update_argument",
    "cbor_encoder",
    "cbor_decoder",
    "nme_cbor_encoder",
    "nme_cbor_decoder",
    "__version__",
)



import pytest


@pytest.fixture()
def clean_register():
    from local_migrator._serialize_hooks import REGISTER

    def clean():
        REGISTER._data_dkt = {}

    old_dict = REGISTER._data_dkt
    clean()
    yield clean
    REGISTER._data_dkt = old_dict



import sys

try:
    import local_migrator  # noqa: F401
except ImportError as e:
    if e.args[0].startswith("local_migrator is incompatible with nme<=0.1.6. You need to upgrade or uninstall nme."):
        print("Ok")
        sys.exit(0)
print("ImportError not raised")
sys.exit(1)



import json

from pydantic import BaseModel

from local_migrator import Encoder, object_hook


class SampleModel(BaseModel):
    field1: int
    field2: str


data = SampleModel(field1=4, field2="abc")

with open("sample.json", "w") as f_p:
    json.dump(data, f_p, cls=Encoder)

with open("sample.json") as f_p:
    data2 = json.load(f_p, object_hook=object_hook)

assert data == data2



from enum import Enum
from typing import Any, Dict

import pytest

from local_migrator import REGISTER, class_to_str, register_class, rename_key, update_argument


@register_class
class SampleClass1:
    pass


def rename_a_to_c(dkt: Dict[str, Any]) -> Dict[str, Any]:
    dkt = dict(dkt)
    dkt["c"] = dkt["a"]
    del dkt["a"]
    return dkt


@register_class(version="0.0.1", migrations=[("0.0.1", rename_a_to_c)])
class SampleClass2:
    pass


class SampleClass3:
    pass


@register_class(old_paths=["test.test.BBase"], version="0.0.2")
class SampleClass4:
    pass


class SampleClass6:
    pass


def test_migrate():
    assert REGISTER.migrate_data(class_to_str(SampleClass1), {}, {"a": 1, "b": 2}) == {"a": 1, "b": 2}
    assert REGISTER.migrate_data(
        class_to_str(SampleClass2), {class_to_str(SampleClass2): "0.0.1"}, {"a": 1, "b": 2}
    ) == {"a": 1, "b": 2}
    assert REGISTER.migrate_data(class_to_str(SampleClass2), {}, {"a": 1, "b": 2}) == {"c": 1, "b": 2}


def test_unregistered_class():
    assert REGISTER.get_class(class_to_str(SampleClass3)) is SampleClass3


def test_old_paths():
    assert REGISTER.get_class("test.test.BBase") is SampleClass4


def test_import_part():
    obj = REGISTER.get_class("class_register_util.SampleClass5")
    from class_register_util import SampleClass5

    assert SampleClass5 is obj


def test_old_class_error():
    with pytest.raises(RuntimeError):
        register_class(SampleClass6, old_paths=[class_to_str(SampleClass4)])


def test_get_version():
    assert str(REGISTER.get_version(SampleClass1)) == "0.0.0"
    assert str(REGISTER.get_version(SampleClass2)) == "0.0.1"
    assert str(REGISTER.get_version(SampleClass4)) == "0.0.2"


def test_rename_key():
    dkt = {"aaa": 1, "bbb": 2}
    assert rename_key(from_key="aaa", to_key="ccc")(dkt) == {"bbb": 2, "ccc": 1}
    with pytest.raises(KeyError):
        rename_key("ccc", "ddd")(dkt)

    assert rename_key("ccc", "ddd", optional=True)(dkt) == dkt


def test_update_argument(clean_register):
    @REGISTER.register(version="0.0.1")
    class MigrateClass:
        def __init__(self, a, b):
            self.a = a
            self.b = b

    class ClassToCall:
        __argument_class__ = MigrateClass

        @classmethod
        @update_argument("arg")
        def call_func(cls, aa, arg: MigrateClass):
            print(aa, arg.a)

    ClassToCall.call_func(aa=1, arg={"a": 1, "b": 2})
    ClassToCall.call_func(1, {"a": 1, "b": 2})


def test_migrate_parent_class(clean_register):
    @register_class(version="0.0.1", migrations=[("0.0.1", rename_key("field", "field1"))])
    class BaseMigrateClass:
        def __init__(self, field1):
            self.field1 = field1

    @register_class
    class MigrateClass(BaseMigrateClass):
        def __init__(self, field2, **kwargs):
            super().__init__(**kwargs)
            self.field2 = field2

    migrated = REGISTER.migrate_data(class_to_str(MigrateClass), {}, {"field2": 1, "field": 5})
    assert "field1" in migrated


def test_migrate_by_class(clean_register):
    @register_class(version="0.0.1", migrations=[("0.0.1", rename_key("field", "field1"))])
    class MigrateClass:
        def __init__(self, field1):
            self.field1 = field1

    migrated = REGISTER.migrate_data(class_to_str(MigrateClass), {}, {"field": 5})
    migrated2 = REGISTER.migrate_data(MigrateClass, {}, {"field": 5})
    assert migrated == {"field1": 5}
    assert migrated == migrated2


def test_migrate_keywords_args(clean_register):
    @register_class(version="0.0.1", migrations=[("0.0.1", rename_key("field", "field1"))])
    class MigrateClass:
        def __init__(self, field1):
            self.field1 = field1

    migrated = REGISTER.migrate_data(cls=class_to_str(MigrateClass), class_str_to_version_dkt={}, data={"field": 5})
    with pytest.warns(FutureWarning):
        migrated2 = REGISTER.migrate_data(
            class_str=class_to_str(MigrateClass), class_str_to_version_dkt={}, data={"field": 5}
        )
    assert migrated == {"field1": 5}
    assert migrated == migrated2


def test_wrong_version(clean_register):
    class SampleEnum(Enum):
        value1 = 1

    with pytest.raises(ValueError, match="class version lower than in migrations"):
        register_class("0.0.0", migrations=[("0.0.1", lambda x: x)])(SampleEnum)


def test_double_register(clean_register):
    class SampleEnum(Enum):
        value1 = 1

    register_class(SampleEnum)

    with pytest.raises(RuntimeError):
        register_class(SampleEnum)


def test_missed_class(clean_register):
    with pytest.raises(ValueError, match=r"Class [\w\.]+ not found"):
        REGISTER.get_class("nme.not_package.not_module.NotClass")
    with pytest.raises(ValueError, match=r"Class [\w\.]+ not found"):
        REGISTER.get_class("nme_not.not_package.not_module.NotClass")


def test_allow_errors(clean_register):
    @register_class
    class _SampleClass1:
        pass

    @register_class(allow_errors_in_values=True)
    class _SampleClass2:
        pass

    assert not REGISTER.allow_errors_in_values(_SampleClass1)
    assert REGISTER.allow_errors_in_values(_SampleClass2)
